import re
from dataclasses import dataclass, field
from typing import Any, Callable, Optional

import mldesigner as mld
from azure.ai.ml.dsl import pipeline as pipeline_dec
from azure.ai.ml.entities import Environment
from typing_extensions import override

from ez_azml.cloud_runs.cloud_run import CloudRun, RunOutput


@dataclass
class PipelineCommand:
    """Class representing a mldesigner Azureml step (mldesigner.command_component).

    Args:
        function: function to be decorated
        environment: enviroment to run the comand at
        component_kwargs: kwargs to pass to the `mldesigner.command_component` decorator
    """

    function: Callable
    environment: Optional[Environment] = None
    component_kwargs: Optional[dict[str, Any]] = field(default_factory=dict)

    def __post_init__(self):
        self.component_kwargs["environment"] = self.environment

    @property
    def name(self):
        """Command's name."""
        return self.function.__name__


class Pipeline(CloudRun):
    """Cloud run that uses mldesigner pipelines.

    Args:
        experiment_name: experiment this pipeline is associated with
        commands: commands to use within the pipeline
        pipeline: function representing the pipeline
        dec_kwargs: kwargs for azure.ai.ml.dsl.pipeline decorator
    """

    def __init__(
        self,
        experiment_name: str,
        commands: list[PipelineCommand],
        pipeline: Callable,
        dec_kwargs: Optional[dict[str, Any]] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.experiment_name = experiment_name
        self.commands = commands
        self.dec_kwargs = dec_kwargs or {}
        self.pipeline = pipeline

    def _register_components(
        self, commands: list[PipelineCommand]
    ) -> list[PipelineCommand]:
        kwargs_pattern = r"(\w+)\s*=\s*(['\"]?)(\w+)\2"
        for command in commands:
            fn = command.function
            for parameter in fn.__annotations__:
                hint = fn.__annotations__[parameter]
                if isinstance(hint, str):
                    potential_kwargs = {
                        key: value for key, _, value in re.findall(kwargs_pattern, hint)
                    }
                    if "Input(" in hint:
                        hint = mld.Input(**potential_kwargs)
                    elif "Output(" in hint:
                        hint = mld.Output(**potential_kwargs)
                    fn.__annotations__[parameter] = hint
            command.function = mld.command_component(**command.component_kwargs)(fn)
        return commands

    def _build_pipeline(self, pipeline: Callable, dec_kwargs: Optional[dict[str, Any]]):
        # Inject the decorated functions
        for f in self.commands:
            pipeline.__globals__[f.__name__] = f
        f = pipeline_dec(**dec_kwargs)(pipeline)
        return pipeline_dec(**dec_kwargs)(pipeline)

    def _setup_dec_kwargs(self, dec_kwargs: dict[str, Any]):
        if self.compute:
            dec_kwargs["default_compute"] = dec_kwargs.get(
                "default_compute", self.compute.name
            )
        if self.environment:
            dec_kwargs["default_compute"] = dec_kwargs.get(
                "default_compute", self.compute.name
            )
        return dec_kwargs

    @override
    def on_run_start(self):
        self.commands = self._register_components(self.commands)
        self.dec_kwargs = self._setup_dec_kwargs(self.dec_kwargs)
        self.pipeline = self._build_pipeline(self.pipeline, self.dec_kwargs)

    @override
    def run(self) -> RunOutput:
        pipeline_job = self.pipeline(**self.inputs, **self.outputs)
        pipe = self.ml_client.jobs.create_or_update(
            pipeline_job, experiment_name=self.experiment_name
        )
        return RunOutput(url=pipe.studio_url)



import inspect
from typing import Callable
from unittest import mock

from ez_azml.cloud_runs.pipelines import Pipeline


def assert_function_wrapped(function: Callable, wrapped_function: Callable) -> None:
    """Asserts that a function is properly wrapped.

    Args:
        wrapped_function: The function that is potentially wrapped.
        function: The original function to compare against after unwrapping.

    Raises:
        AssertionError: If the function is already wrapped or if unwrapping the wrapped
                        function does not yield the original function.
    """
    assert not hasattr(function, "__wrapped__"), "The function should not be wrapped."
    assert (
        inspect.unwrap(wrapped_function) is function
    ), "Unwrapping the wrapped function did not yield the original function."


def test_pipeline_register_components(pipeline: Pipeline):
    """Tests that the components are properly decorated."""
    raw_functions = [c.function for c in pipeline.commands]
    registered_commands = pipeline._register_components(pipeline.commands)
    for raw_function, registered in zip(raw_functions, registered_commands):
        assert_function_wrapped(raw_function, registered.function)


def test_pipeline_setup_dec_kwargs(pipeline: Pipeline):
    """Tests that the pipeline decorator kwargs are properly postprocessed."""
    pipeline.dec_kwargs = pipeline._setup_dec_kwargs(pipeline.dec_kwargs)


def test_build_pipeline(pipeline: Pipeline):
    """Tests that the pipeline is properly decorated."""

    def f():
        pass

    with mock.patch.object(pipeline, "commands", return_value=f) as mocked:
        mocked.__name__ = "command_fn"

        raw_pipeline = pipeline.pipeline
        built_pipeline = pipeline._build_pipeline(raw_pipeline, pipeline.dec_kwargs)
        assert_function_wrapped(raw_pipeline, built_pipeline)



from typing import Callable

import mldesigner as mld
import pytest
from azure.ai.ml import Input, Output
from ez_azml.cloud_runs.pipelines import Pipeline, PipelineCommand


@pytest.fixture()
def command_fn() -> Callable:
    """Function used in mldesigner command component."""

    def command__fn(
        command_input: mld.Input(type="uri_folder"),  # type: ignore
        command_output: mld.Output(type="uri_folder"),  # type: ignore
    ):
        print("this is a test_fn")

    return command__fn


@pytest.fixture()
def pipeline_fn(command_fn) -> Callable:
    """Function used in azure.ai.ml pipelines."""

    def pipeline__fn(
        pipeline_input: Input,
        pipeline_output: Output,
    ):
        command_fn(pipeline_input)

    return pipeline__fn


@pytest.fixture()
def pipeline_command(command_fn: Callable):
    """PipelineCommand fixture."""
    return PipelineCommand(function=command_fn)


@pytest.fixture()
def pipeline(pipeline_fn: Callable, pipeline_command: PipelineCommand):
    """Pipeline cloud run fixture."""
    return Pipeline(
        experiment_name="fixture",
        commands=[pipeline_command],
        pipeline=pipeline_fn,
    )



from typing import Optional

from azure.ai.ml import command
from azure.ai.ml.entities import (
    UserIdentityConfiguration,
    WorkspaceConnection,
)
from typing_extensions import override

from ez_azml.cloud_runs.cloud_run import CloudRun, RunOutput


class Command(CloudRun):
    """Cloud run that is based on AzureML Commands.

    Args:
        code: location of the scripts to use
        commands: commands to run on the cloud (e.g. `python my_script.py`)
        flags: flags to use with the last command.
        identity: credentials to use.
    """

    def __init__(
        self,
        code: str,
        commands: list[str],
        ws_connection: Optional[WorkspaceConnection] = None,
        flags: Optional[list[str]] = None,
        identity: Optional[UserIdentityConfiguration] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.ws_connection = ws_connection
        identity = identity or UserIdentityConfiguration()
        if flags:
            commands[-1] += " " + " ".join(flags)
        self.job = command(
            command=";".join(commands),
            code=code,
            environment=self.environment,
            compute=self.compute.name,
            inputs=self.inputs,
            outputs=self.outputs,
            identity=identity,
        )

    @override
    def run(self) -> str:
        if self.ws_connection:
            self.ml_client.connections.create_or_update(self.ws_connection)
        self.ml_client.environments.create_or_update(self.environment)
        self.ml_client.begin_create_or_update(self.compute).result()
        cloud_job = self.ml_client.create_or_update(self.job)
        return RunOutput(url=cloud_job.studio_url)



# %%
import os
from typing import Any, Optional

from azure.ai.ml import MLClient
from azure.core.credentials import TokenCredential
from azure.identity import DefaultAzureCredential


class EzMLClient(MLClient):
    """A easy client class to interact with Azure ML services.

    Args:
        credential: The credential to use for authentication.
        Defaults to DefaultAzureCredential()
        subscription_id: The Azure subscription ID.
        Defaults to env AZURE_SUBSCRIPTION_ID.
        resource_group_name: The Azure resource group.
        Defaults to env AZURE_RESOURCE_GROUP_NAME.
        workspace_name: The workspace to use in the client.
        Defaults to env AZURE_WORKSPACE.
    """

    def __init__(
        self,
        credential: Optional[TokenCredential] = None,
        subscription_id: Optional[str] = None,
        resource_group_name: Optional[str] = None,
        workspace_name: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        credential = credential or DefaultAzureCredential()
        subscription_id = subscription_id or os.getenv("AZURE_SUBSCRIPTION_ID")
        resource_group_name = resource_group_name or os.getenv(
            "AZURE_RESOURCE_GROUP_NAME"
        )
        workspace_name = workspace_name or os.getenv("AZURE_WORKSPACE")
        super().__init__(
            credential, subscription_id, resource_group_name, workspace_name, **kwargs
        )
